@@ -54,7 +54,7 @@ use Coro::ProcessPool::Util;
 use Coro::Semaphore;
 require Coro::ProcessPool::Pipeline;
 
-our $VERSION = '0.25_1';
+our $VERSION = '0.25_2';
 
 if ($^O eq 'MSWin32') {
     die 'MSWin32 is not supported';
@@ -122,13 +122,13 @@ has num_procs => (
 
 =head2 procs
 
-Array holding the L<Coro::Process> objects.
+Array holding the L<Coro::ProcessPool::Process> objects.
 
 =cut
 
 has procs => (
     is      => 'ro',
-    isa     => ArrayRef[InstanceOf['Coro::Process']],
+    isa     => ArrayRef[InstanceOf['Coro::ProcessPool::Process']],
     default => sub { [] },
 );
 
@@ -138,7 +138,7 @@ has procs => (
 
 has all_procs => (
     is      => 'ro',
-    isa     => Map[Str, InstanceOf['Coro::Process']],
+    isa     => HashRef[InstanceOf['Coro::ProcessPool::Process']],
     default => sub { {} },
 );
 
@@ -169,7 +169,7 @@ sub DEMOLISH {
 sub BUILD {
     my $self = shift;
     for (1 .. $self->max_procs) {
-        $self->checkin_proc($self->start_proc);
+        unshift @{$self->procs}, $self->start_proc;
     }
 }
 
@@ -193,7 +193,12 @@ sub kill_proc {
 sub checkin_proc {
     my ($self, $proc) = @_;
 
-    if (!$self->is_running || ($self->max_reqs && $proc->messages_sent >= $self->max_reqs)) {
+    unless ($self->is_running) {
+      $self->kill_proc($proc);
+      return;
+    }
+
+    if ($self->max_reqs && $proc->messages_sent >= $self->max_reqs) {
         $self->kill_proc($proc);
         unshift @{$self->procs}, $self->start_proc;
     } else {
@@ -244,6 +249,9 @@ sub shutdown {
     $self->{procs}      = [];
     $self->{all_proces} = {};
     $self->{num_procs}  = 0;
+    $self->{procs_lock} = Coro::Semaphore->new($self->max_procs);
+
+    return;
 }
 
 =head2 process($f, $args)
