@@ -32,9 +32,14 @@ sub BUILDARGS {
     my $exec = "$cmd $args";
     my $pid  = open3($w, $r, $e, $exec) or croak "Error spawning process: $!";
 
-    $SIG{CHLD} = 'IGNORE';
+    # Get real child process id (open3 creates parent -> sh -> perl). Worker
+    # will send its own pid once initialized.
+    my $child = <$r>;
+    $child or croak 'unable to talk to child process';
+    chomp $child;
 
     $args{pid}       = $pid;
+    $args{child}     = $child;
     $args{child_in}  = unblock $r;
     $args{child_out} = unblock $w;
     $args{child_err} = unblock $e;
@@ -54,6 +59,8 @@ sub DEMOLISH {
     $self->shutdown;
 }
 
+has child => (is => 'rw', isa => Int);
+
 has messages_sent => (
     is       => 'rw',
     isa      => Int,
@@ -92,16 +99,10 @@ has child_err => (
 
 has inbox => (
     is       => 'ro',
-    isa      => HashRef[Str, InstanceOf['Coro::Channel']],
+    isa      => HashRef,
     default  => sub { {} },
 );
 
-has on_read => (
-    is      => 'rw',
-    isa     => Maybe[CodeRef],
-    clearer => 'clear_on_read',
-);
-
 has child_in_watcher => (
     is       => 'lazy',
     isa      => InstanceOf['Coro'],
@@ -114,16 +115,10 @@ sub _build_child_in_watcher {
         my $self = shift;
 
         while (my $line = $self->child_in->readline($EOL)) {
-            if ($self->on_read) {
-                $self->on_read->();
-                $self->clear_on_read;
-            }
-
-            my $msg  = decode($line);
-            my ($id, $data) = @$msg;
+            my ($id, $error, $data) = decode($line);
 
             if (exists $self->inbox->{$id}) {
-                $self->inbox->{$id}->send($data);
+                $self->inbox->{$id}->send([$error, $data]);
             } else {
                 warn "Unexpected message received: $id";
             }
@@ -143,6 +138,7 @@ sub _build_child_err_watcher {
         my $self = shift;
         while (my $line = $self->child_err->readline($EOL)) {
             warn sprintf("(WORKER PID %s) %s", ($self->pid || '(DEAD)'), $line);
+            last unless $self->pid;
         }
     } @_;
 }
@@ -156,18 +152,22 @@ sub cleanup {
     }
 
     if ($self->child_in) {
-        $self->child_in_watcher->join;
+        $self->child_in_watcher->safe_cancel;
         $self->child_in->close;
         $self->clear_child_in_watcher;
         $self->clear_child_in;
     }
 
     if ($self->child_err) {
+        $self->child_err_watcher->safe_cancel;
         $self->child_err->close;
-        $self->child_err_watcher->join;
         $self->clear_child_err_watcher;
         $self->clear_child_err;
     }
+
+    foreach my $id (keys %{$self->inbox}) {
+        $self->inbox->{$id}->croak('process killed while waiting on this task to complete');
+    }
 }
 
 sub join {
@@ -194,40 +194,33 @@ sub join {
 
 sub kill_process {
     my ($self, $timeout) = @_;
-
     return unless $self->is_running;
-
-    my $id = $self->write('SHUTDOWN');
-    my $reply = $self->recv($id);
-    $self->child_out->close;
-
-    until ($self->join($timeout)) {
-        kill('KILL', $self->pid);
-        waitpid($self->pid, 0);
-    }
+    kill('KILL', $self->child);
+    waitpid($self->pid, 0);
 }
 
 sub shutdown {
     my ($self, $timeout) = @_;
+    return unless $self->pid;
     $self->kill_process($timeout);
     $self->cleanup;
     return 1;
 }
 
 sub write {
-    my ($self, $data) = @_;
+    my ($self, $task, $args) = @_;
+    croak 'not running' unless $self->is_running;
+    croak 'process disconnected' unless $self->child_out;
     my $id = Data::UUID->new->create_str();
     $self->inbox->{$id} = AnyEvent->condvar;
-    $self->child_out->print(encode([$id, $data]) . $EOL);
+    $self->child_out->print(encode($id, $task, $args) . $EOL);
     ++$self->{messages_sent};
     return $id;
 }
 
 sub send {
     my ($self, $f, $args) = @_;
-    croak 'not running' unless $self->is_running;
-    $args ||= [];
-    return $self->write([$f, $args]);
+    return $self->write($f, $args || []);
 }
 
 sub recv {
